#!/usr/bin/env Rscript
args <- commandArgs(trailingOnly=T)

if (length(args) != 3 || args[1] == "-h" || args[1] == "--help") {
  cat("\n")
  cat("Extracts objects with a high similarity to a given object or list of objects.\n")
  cat("The result is printed to stdout.\n")
  cat("\n")
  cat("Required arguments\n")
  cat("--------------------\n")
  cat("dataset - string: assumes that files dataset.matrix.gz and dataset.class exist in the current working directory\n")
  cat("objects - list of strings: identifiers of the objects for which similar objects should be extracted, e.g., 'MABC MXYZ'\n")
  cat("threshold - numeric, >=0 and <=1: similarity threshold for objects to be extracted\n")
  quit(save="no", status=0)
}

data.set.name <- args[1]
object.names <- args[2]
object.names <- strsplit(object.names, ' ', fixed=T)[[1]]
sim.t <- as.numeric(args[3])

cat(paste("Scanning", data.set.name, "for objects with similarities >=", sim.t, "to the objects:\n"), file=stderr())
cat(paste(paste(object.names, collapse=" "), "\n"), file=stderr())

sim.df <- read.delim(gzfile(paste0(data.set.name, '.matrix.gz')), header=F, sep=" ")
sim.df <- sim.df[, -ncol(sim.df)]
sample.names <- read.delim(paste0(data.set.name, '.class'), header=F, stringsAsFactors=F)
sample.names <- as.character(sample.names[,1])

similarity.matrix <- as.matrix(sim.df)
rownames(similarity.matrix) <- sample.names
colnames(similarity.matrix) <- sample.names

#############
# FUNCTIONS
#############
extract.similar.candidates <- function(sim.matrix, candidate.names, sim.threshold) {
  dimensions <- dim(sim.matrix)
  if (dimensions[1] != dimensions[2]) {
    stop("Given similarity matrix must be symmetric.")
  }
  
  rnames <- rownames(sim.matrix)
  cnames <- colnames(sim.matrix)
  if (is.null(rnames)) {
    stop("Row names of given similarity matrix must not be NULL.")
  } else if (is.null(cnames)) {
    stop("Column names of given similarity matrix must not be NULL.")
  }
  
  if (!all(rnames == cnames)) {
    stop("Row and column names must be identical.")
  }
  
  if (length(unique(rnames)) != length(rnames)) {
    stop("Row and column names must ne unique.")
  }
  
  if (sim.threshold < 0 || sim.threshold > 1) {
    stop("Similarity threshold must be >= 0 and <= 1.")
  }
  
  row.indices <- sapply(candidate.names, function(cand.name) {
    cand.index <- which(rnames == cand.name)
    if (length(cand.index) == 0) {
      stop(paste("Candidate name", cand.name, "was not found in similarity matrix."))
    }
    cand.index
  })
  
  similar.motif.list <- lapply(1:length(candidate.names), function(cand.index) {
    cand.name <- candidate.names[cand.index]
    row.index <- row.indices[cand.index]
    sim.vector <- sim.matrix[row.index, ]
    similar.motif.indices <- which(sim.vector >= sim.threshold)
    similar.motif.names <- rnames[similar.motif.indices]
    similarity.values <- sim.vector[similar.motif.indices]
    data.frame(Motif=similar.motif.names, Sim=similarity.values)
  })
  names(similar.motif.list) <- candidate.names
  similar.motif.list
}

###########
# Execution code
###########
sim.cand.list <- extract.similar.candidates(similarity.matrix, object.names, sim.t)
cat("Query-Object\tSimilar-Object\tSimilarity\n")
for (i in 1:length(sim.cand.list)) {
  query.obj <- names(sim.cand.list)[i]
  curr.sim.df <- sim.cand.list[[i]]
  for (j in 1:nrow(curr.sim.df)) {
    cat(query.obj)
    cat("\t")
    cat(curr.sim.df[j,1])
    cat("\t")
    cat(curr.sim.df[j,2])
    cat("\n")
  }
}