#!/usr/bin/env Rscript
suppressMessages(library('ggplot2'))
suppressMessages(library('cluster')) # silhouette value

args <- commandArgs(trailingOnly=T)

if (length(args) != 2 || args[1] == "-h" || args[1] == "--help") {
  cat("\n")
  cat("Performs hierarchical clustering based on an EDeN similarity matrix.\n")
  cat("Agglomerative with all linkage methods available in R is performed.\n")
  cat("The silhouette value is used to choose the optimal number of clusters.\n")
  cat("Also produces various plots visualizing these results:\n")
  cat("- distribution of similarity values\n")
  cat("- multidimensional scaling to two dimensions\n")
  cat("- plots EDeN similarity of two alignments over the difference of their average sequence lengths\n")
  cat("\n")
  cat("Required arguments\n")
  cat("--------------------\n")
  cat("dataset - string: assumes that files dataset.matrix.gz and dataset.class exist in working directory\n")
  cat("outdir - string: path to the output directory\n")
  cat("\n")
  quit(save="no", status=0)
}
data.set.name <- args[1]
output.dir <- args[2]
# data.set.name <- 'UC-CON-all-motifs-logfoldchange-gt-2'
# output.dir <- 'R-output'

max.cluster.count <- 20

sim.df <- read.delim(gzfile(paste0(data.set.name, '.matrix.gz')), header=F, sep=" ")
sim.df <- sim.df[, -ncol(sim.df)]
sample.names <- read.delim(paste0(data.set.name, '.class'), header=F, stringsAsFactors=F)
sample.names <- as.character(sample.names[,1])

sim.matrix <- as.matrix(sim.df)
rownames(sim.matrix) <- sample.names
colnames(sim.matrix) <- sample.names

sim.df <- data.frame(sim.matrix)
sim.df <- cbind.data.frame(Motif=rownames(sim.matrix), sim.df)
write.table(sim.df,
            file.path(output.dir,paste0("sim-matrix.tsv")),
            sep="\t", row.names=F,
            col.names=c("Motif", paste(colnames(sim.df)[-1], data.set.name, sep="-")),
            quote=F)

# TODO this is not used so far, add accessibility analysis
rbp.binding.type <- read.delim('/home/projects/rth_roi/data/RBP/rbp_domain.tsv.gz', header=F)

#three.d.motifs <- read.delim('/home/projects/users/corinna/CMfinder_Screen_complete/cmfIDs_threshcand_occ50_all.txt', header=F)
#three.d.motifs.df <- read.delim('/home/projects/users/corinna/CMfinder_Screen_complete/cmmotifs_cluster50.txt', sep=" ", header=F)
three.d.motifs.df <- read.delim('/home/projects/users/corinna/CMfinder_Screen_complete/cmmotifs_cluster50_all.txt.gz',
								sep=" ", header=F)
three.d.motifs.df <- three.d.motifs.df[, c(14,4,13)]
colnames(three.d.motifs.df) <- c("CMFmot", "3Dmod", "highConfidence")

overlapping.motifs.df <- three.d.motifs.df[as.character(three.d.motifs.df[,1]) %in% sample.names, ]
write.table(overlapping.motifs.df,
           file.path(output.dir,'overlapping-with-3D-motifs.txt'),
           sep="\t", row.names=F, col.names=F, quote=F)
overlapping.motifs <- unique(as.character(overlapping.motifs.df[,1]))
overlapping.motifs.sim.matrix <- sim.matrix[overlapping.motifs,overlapping.motifs]

na.ind <- is.na(as.numeric(overlapping.motifs.df[,3]))
overlapping.motifs.df[na.ind, 3] <- 0
overlapping.motifs.high.reliability <- unique(as.character(overlapping.motifs.df[overlapping.motifs.df[,3] > 0, 1]))

sim.dist.file.name.png <- file.path(output.dir,'similarity-density.png')
sim.dist.file.name.pdf <- file.path(output.dir,'similarity-density.pdf')
cat(paste0("Plotting distribution of similarity values to file ",
           sim.dist.file.name.png,"\n"))
sim.vector <- sim.matrix[upper.tri(sim.matrix)]
plot.df <- data.frame(Similarity=sim.vector)
m <- ggplot(plot.df, aes(x=Similarity))
m <- m + geom_density(adjust=0.1) +
  labs(title='Distribution of pairwise similarities computed by EDeN')
png(sim.dist.file.name.png, width=1400, height=700)
print(m)
dev.off()
pdf(sim.dist.file.name.pdf, width=10, height=5)
print(m)
dev.off()
rm(m)
rm(plot.df)
cat("Done.\n\n")

cat("Started converting similarity matrix to dissimilarity matrix.\n")
# Convert a similarity matrix into dissimilarity matrix by multiplying all
# values by -1 and then addding the max. similarity to every value to obtain
# a matrix of positive values.
max.sim <- max(sim.matrix)
dist.matrix <- as.dist((-sim.matrix) + max.sim)
cat("Done.\n\n")

mds.dim.2.matrix <- cmdscale(dist.matrix, k=2)
plot.df <- data.frame(x=mds.dim.2.matrix[,1], y=mds.dim.2.matrix[,2])
p <- ggplot(plot.df, aes(x, y)) + geom_point(shape=1, alpha=1)
p <- p + labs(title="Multi-dimensional scaling to 2 dimensions")
png(file.path(output.dir,'mds-two-dim.png'), width=750, height=750)
print(p)
dev.off()
pdf(file.path(output.dir,'mds-two-dim.pdf'), width=8, height=7)
print(p)
dev.off()

best.k.overall <- 0
best.linkage.overall <- ""
best.sil.overall <- -2
all.clusterings.df <- data.frame(Motif=sample.names)
for (linkage.method in
     c("ward.D", "ward.D2", "single", "complete",
       "average", "mcquitty", "median", "centroid")) {
  cat(paste0("Started hierarchical clustering with linkage method ",
             linkage.method, "\n"))
  hclust.res <- hclust(dist.matrix, method=linkage.method)
  hclust.file <- file.path(output.dir,paste0("hclust-", linkage.method,"-result.RData.xz"))
  save(hclust.res, file=hclust.file, compress="xz")
  
  png(file.path(output.dir,paste0("hclust-", linkage.method,"-result.png")), width=1500, height=750)
  plot(hclust.res, labels=F)
  dev.off()
  pdf(file.path(output.dir,paste0("hclust-", linkage.method,"-result.pdf")), width=10, height=5)
  plot(hclust.res, labels=F)
  dev.off()
  if (length(overlapping.motifs) > 0) {
    suppressMessages(library('dendextend'))
    dendro <- as.dendrogram(hclust.res)
    highlight.labels <- labels(dendro) %in% overlapping.motifs
    highlight.labels <- as.numeric(highlight.labels)
    png(file.path(output.dir,paste0("hclust-", linkage.method,"-highlighted-overlapping-result.png")), width=1500, height=750)
    plot(hclust.res, labels=F)
    colored_bars(colors = highlight.labels, dend = dendro, y_scale=20)
    dev.off()
    pdf(file.path(output.dir,paste0("hclust-", linkage.method,"-highlighted-overlapping-result.pdf")), width=10, height=5)
    plot(hclust.res, labels=F)
    colored_bars(colors = highlight.labels, dend = dendro, y_scale=20)
    dev.off()
  }  
  cat("Done.\n\n")
  
  cat("Using silhouette value to determine the optimal number of clusters.\n")
  max.k <- min(max.cluster.count, nrow(sim.matrix))
  best.k <- 0
  best.sil <- -2
  all.sil <- c()
  for (curr.k in (2:max.k)) {
    # curr.k <- 5
    clusters <- cutree(hclust.res, k=curr.k)
    sil <- silhouette(clusters, dist.matrix)
    median.sil.val <- median(sil[,3])
    all.sil <- c(all.sil, median.sil.val)
    if (median.sil.val > best.sil) {
      best.sil <- median.sil.val
      best.k <- curr.k
    }
    if (median.sil.val > best.sil.overall) {
      best.k.overall <- curr.k
      best.linkage.overall <- linkage.method
      best.sil.overall <- median.sil.val
    }
  }
  cat(paste0("Optimal number of cluster was found to be: ", best.k, "\n"))
  cat("Done.\n\n")
  sil.plot.df <- data.frame(NumberOfClusters=2:max.k, AverageSilhouetteValue=all.sil)
  write.table(sil.plot.df,
              file.path(output.dir,paste0("hclust-", linkage.method,"-silhouettes.tsv")),
              sep="\t", row.names=F, quote=F)
  
  p <- ggplot(sil.plot.df, aes(x=NumberOfClusters, y=AverageSilhouetteValue))
  p <- p + geom_line() + labs(x="Number of clusters", y="Average silhouette value")
  
  png(file.path(output.dir,paste0("hclust-", linkage.method,"-silhouettes.png")), width=1500, height=750)
  print(p)
  dev.off()
  pdf(file.path(output.dir,paste0("hclust-", linkage.method,"-silhouettes.pdf")), width=10, height=5)
  print(p)
  dev.off()
  opt.clusters <- cutree(hclust.res, k=best.k)
  opt.clusters.df <- data.frame(names(opt.clusters), opt.clusters)
  colnames(opt.clusters.df) <- c("CMFmotif",
                                 paste0("Optimal-Cluster-HierarchicalClustering-", linkage.method, "-Linkage"))
  write.table(opt.clusters.df,
              file.path(output.dir,paste0("hclust-", linkage.method, "-optimal-clusters.tsv")),
              sep="\t", row.names=F, quote=F)
  all.clusterings.df <- cbind.data.frame(all.clusterings.df,
                                         opt.clusters[as.character(all.clusterings.df[,1])])
  colnames(all.clusterings.df)[ncol(all.clusterings.df)] <- 
    paste0("Optimal-Cluster-HierarchicalClustering-", linkage.method, "-Linkage")
  
  # TODO mark optimal clustering in dendrogram
  # TODO highlight clusters in 2D MDS
}
# TODO do the same visualization for overall best clustering

best.hclust.res <- hclust(dist.matrix, method=best.linkage.overall)
best.clusters.overall <- cutree(best.hclust.res, k=best.k.overall)
best.clusters.overall.df <- data.frame(names(best.clusters.overall),
                                       best.clusters.overall)
colnames(best.clusters.overall.df) <- c("CMFmotif",
                               paste0("Optimal-Cluster-HierarchicalClustering-ALL-(", best.linkage.overall, ")"))
write.table(best.clusters.overall.df,
            file.path(output.dir,paste0("hclust-optimal-clusters.tsv")),
            sep="\t", row.names=F, quote=F)

all.clusterings.df <- cbind.data.frame(all.clusterings.df,
                                       best.clusters.overall[as.character(all.clusterings.df[,1])])
colnames(all.clusterings.df)[ncol(all.clusterings.df)] <- 
  paste0("Optimal-Cluster-HierarchicalClustering-ALL-(", best.linkage.overall, ")")
write.table(all.clusterings.df,
            file.path(output.dir,paste0("hclust-all-clusters.tsv")),
            sep="\t", row.names=F, col.names=c("Motif", paste(colnames(all.clusterings.df)[-1], data.set.name, sep="-")),quote=F)
  

# If motifs with 3D structures have been found, draw an 2D MDS with these objects highlighted 
if (length(overlapping.motifs) > 0) {
  cat.vec <- sample.names %in% overlapping.motifs
  cat.vec <- sapply(cat.vec, function(val) if(val) {'3Dcluster'} else {'None'})
  high.reliab <- sample.names %in% overlapping.motifs.high.reliability
  cat.vec[high.reliab] <- '3Dcandidate'
  plot.df <- cbind.data.frame(plot.df, Module=factor(cat.vec))
  name.to.expression <- c(expression('3D'['cluster']), expression('3D'['candidate']), expression('None'))
  names(name.to.expression) <- c('3Dcluster', '3Dcandidate', 'None')
  p <- ggplot(plot.df, aes(x, y, color=Module, shape=Module)) + geom_point(alpha=1) + scale_shape(solid = FALSE) 
  p <- p + labs(title="Multi-dimensional scaling to 2 dimensions") +
    scale_fill_discrete(labels=name.to.expression[levels(plot.df[['Module']])])
  png(file.path(output.dir,'mds-two-dim-overlap-3D.png'), width=750, height=750)
  print(p)
  dev.off()
  pdf(file.path(output.dir,'mds-two-dim-overlap-3D.pdf'), width=8, height=7)
  print(p)
  dev.off()  
}

rm(plot.df)
rm(hclust.res)
rm(dist.matrix)

# TODO
# Might be created by:
# ~/bin/fasta2average-sequence-length.py -d ../../../../data/20141022-HNRPL-overlap/ > file.path(output.dir,average-sequence-length)
if (file.exists(file.path(output.dir,'average-sequence-length.txt'))) {
  cat(paste0("Started plotting similarity vs difference in sequence length.\n"))
  seq.length.df <- read.delim('average-sequence-length.txt', header=F)
  seq.length <- seq.length.df[, 2]
  names(seq.length) <- as.character(seq.length.df[, 1])
  names(seq.length) <- sapply(strsplit(names(seq.length), ".", fixed=T), "[", 1)
  # Make sure that same order as in .class file
  seq.length <- seq.length[sample.names]
  
  length.dist.matrix <- sapply(seq.length, function(val) abs(seq.length - val))
  
  sim.vec <- sim.matrix[upper.tri(sim.matrix)]
  length.dist.vec <- length.dist.matrix[upper.tri(length.dist.matrix)]
  
  # Plot each pair of entries in similarity and distance matrix
  plot.df <- data.frame(DiffSeqLen=length.dist.vec, Sim=sim.vec)
  p <- ggplot(plot.df, aes(DiffSeqLen, Sim)) + geom_point(shape=1, alpha=0.1) + geom_smooth() 
  p <- p + labs(title="Similarity vs difference in average sequence length for each pair of alignments",
                x="Difference in average sequence length", y="Pairwise similarity")
  png(file.path(output.dir,'similarity-diff-seq-length.png'), width=1400, height=700)
  print(p)
  dev.off()
  pdf(file.path(output.dir,'similarity-diff-seq-length.pdf'), width=10, height=5)
  print(p)
  dev.off()
  cat("Done.\n\n")
  rm(seq.length.df, seq.length, length.dist.matrix, sim.vec, length.dist.vec, plot.df, p)
}

# Extract quantiles from vector of all similarities
rounding.digits <- 5
quantile.probablilties <- seq(0, 1, by=0.01)
quantile.values <- round(quantile(sim.vector, quantile.probablilties),
                         rounding.digits)
rounded.sim.matrix <- round(sim.matrix, rounding.digits)
quantile.pairs <- sapply(quantile.values, function(q.val) {
  hit.pairs <- which(rounded.sim.matrix == q.val, arr.ind = TRUE)
  if (length(hit.pairs) >= 2) {
    hit.pair.id1 <- sample.names[hit.pairs[1,1]]
    hit.pair.id2 <- sample.names[hit.pairs[1,2]]
    c(hit.pair.id1, hit.pair.id2)
  } else {
    c("", "")
  }
})
quantile.pairs <- t(quantile.pairs)
quantile.df <- data.frame(Quantile=names(quantile.values),
                          Similarity=quantile.values,
                          First=quantile.pairs[, 1],
                          Second=quantile.pairs[, 2])
quantile.df[,1] <- as.character(quantile.df[,1])
quantile.df[,3] <- as.character(quantile.df[,3])
quantile.df[,4] <- as.character(quantile.df[,4])
quantile.pairs.file <- 'quantile-pairs.tsv'
write.table(quantile.df, file.path(output.dir,quantile.pairs.file), sep="\t",
            row.names=F, quote=F)
for (row.idx in 1:nrow(quantile.df)) {
  # row.idx <- 1 
  quant <- paste0(formatC(as.numeric(strsplit(quantile.df[row.idx, 1], '%')[[1]]),
                   width = 3, format = "d", flag = "0"), '%')
  sim <- quantile.df[row.idx, 2]
  first.obj <- quantile.df[row.idx, 3]
  second.obj <- quantile.df[row.idx, 4]
  cat(paste0('Processing row ', row.idx, ' with motifs ', first.obj,
             ' and ', second.obj, '.\n'))
  both.objs <- c(first.obj, second.obj)
  for (obj.idx in 1:length(both.objs)) {
    curr.obj <- both.objs[obj.idx]
    cat(paste0('Extracting alignment and structure for ', curr.obj, '.\n'))
    system(paste0('export PATH=$PATH:/home/users/seemann/bin && ',
                  '/home/users/seemann/projects/CMfinder/cm/scripts/getCMfinderImages.sh ',
                  curr.obj))
    file.rename(from=paste0(curr.obj, '.aln.ps'),
                to=file.path(output.dir, paste0('Quantile-', quant, '-', obj.idx, '-', curr.obj, '.aln.ps')))
    file.rename(from=paste0(curr.obj, '.ss.ps'),
                to=file.path(output.dir, paste0('Quantile-', quant, '-', obj.idx, '-', curr.obj, '.ss.ps')))
  }
}
# Highest scoring pairs
number.of.top.pairs <- 10
top.sim.values <-
  sort(sim.vector, decreasing=T)[1:number.of.top.pairs]
top.sim.pairs <- sapply(top.sim.values, function(sim.val) {
  hit.pairs <- which(sim.matrix == sim.val, arr.ind = TRUE)
  if (length(hit.pairs) >= 2) {
    hit.pair.id1 <- sample.names[hit.pairs[1,1]]
    hit.pair.id2 <- sample.names[hit.pairs[1,2]]
    c(hit.pair.id1, hit.pair.id2)
  } else {
    c("", "")
  }
})
top.sim.pairs <- t(top.sim.pairs)
top.sim.pairs.df <- data.frame(Rank=1:number.of.top.pairs,
                               Similarity=top.sim.values,
                               First=top.sim.pairs[, 1],
                               Second=top.sim.pairs[, 2])
top.sim.pairs.df[,3] <- as.character(top.sim.pairs.df[,3]) 
top.sim.pairs.df[,4] <- as.character(top.sim.pairs.df[,4])
top.sim.pairs.file <- paste0('top-', number.of.top.pairs, '-sim-pairs.tsv')
write.table(top.sim.pairs.df, file.path(output.dir,top.sim.pairs.file), sep="\t",
            row.names=F, quote=F)

for (row.idx in 1:nrow(top.sim.pairs.df)) {
  # row.idx <- 1
  ranking <- formatC(top.sim.pairs.df[row.idx, 1], width = 2, format = "d", flag = "0")
  sim <- top.sim.pairs.df[row.idx, 2]
  first.obj <- top.sim.pairs.df[row.idx, 3]
  second.obj <- top.sim.pairs.df[row.idx, 4]
  cat(paste0('Processing row ', row.idx, ' with motifs ', first.obj,
             ' and ', second.obj, '.\n'))
  both.objs <- c(first.obj, second.obj)
  for (obj.idx in 1:length(both.objs)) {
    curr.obj <- both.objs[obj.idx]
    cat(paste0('Extracting alignment and structure for ', curr.obj, '.\n'))
    system(paste0('export PATH=$PATH:/home/users/seemann/bin && ',
                  '/home/users/seemann/projects/CMfinder/cm/scripts/getCMfinderImages.sh ',
                  curr.obj))
    file.rename(from=paste0(curr.obj, '.aln.ps'),
                to=file.path(output.dir, paste0('Rank-', ranking, '-', obj.idx, '-', curr.obj, '.aln.ps')))
    file.rename(from=paste0(curr.obj, '.ss.ps'),
                to=file.path(output.dir, paste0('Rank-', ranking, '-', obj.idx, '-', curr.obj, '.ss.ps')))
  }
}
# TODO pvclust
